@@ -39,6 +39,7 @@
 
 #define STYLESHEET DATADIR"/yelp/xslt/mal2html.xsl"
 #define MALLARD_NS "http://projectmallard.org/1.0/"
+#define MALLARD_FACET_NS "http://projectmallard.org/facet/1.0/"
 
 typedef enum {
     MALLARD_STATE_BLANK,
@@ -68,6 +69,7 @@ typedef struct {
 
     gchar         *page_title;
     gchar         *page_desc;
+    gchar         *next_page;
 } MallardPageData;
 
 static void           yelp_mallard_document_class_init (YelpMallardDocumentClass *klass);
@@ -75,6 +77,7 @@ static void           yelp_mallard_document_init       (YelpMallardDocument
 static void           yelp_mallard_document_dispose    (GObject                  *object);
 static void           yelp_mallard_document_finalize   (GObject                  *object);
 
+static void           mallard_index             (YelpDocument         *document);
 static gboolean       mallard_request_page      (YelpDocument         *document,
                                                  const gchar          *page_id,
                                                  GCancellable         *cancellable,
@@ -101,6 +104,11 @@ static void           mallard_page_data_info    (MallardPageData      *page_data
 static void           mallard_page_data_run     (MallardPageData      *page_data);
 static void           mallard_page_data_free    (MallardPageData      *page_data);
 
+static const char *   xml_node_get_icon         (xmlNodePtr            node);
+static gboolean       xml_node_is_ns_name       (xmlNodePtr            node,
+                                                 const xmlChar        *ns,
+                                                 const xmlChar        *name);
+
 
 G_DEFINE_TYPE (YelpMallardDocument, yelp_mallard_document, YELP_TYPE_DOCUMENT);
 #define GET_PRIV(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), YELP_TYPE_MALLARD_DOCUMENT, YelpMallardDocumentPrivate))
@@ -113,6 +121,8 @@ struct _YelpMallardDocumentPrivate {
     GMutex        *mutex;
     GThread       *thread;
     gboolean       thread_running;
+    GThread       *index;
+    gboolean       index_running;
     GSList        *pending;
 
     xmlDocPtr      cache;
@@ -134,6 +144,7 @@ yelp_mallard_document_class_init (YelpMallardDocumentClass *klass)
     object_class->finalize = yelp_mallard_document_finalize;
 
     document_class->request_page = mallard_request_page;
+    document_class->index = mallard_index;
 
     g_type_class_add_private (klass, sizeof (YelpMallardDocumentPrivate));
 }
@@ -147,6 +158,7 @@ yelp_mallard_document_init (YelpMallardDocument *mallard)
     priv->mutex = g_mutex_new ();
 
     priv->thread_running = FALSE;
+    priv->index_running = FALSE;
 
     priv->cache = xmlNewDoc (BAD_CAST "1.0");
     priv->cache_ns = xmlNewNs (NULL, BAD_CAST MALLARD_NS, BAD_CAST "mal");
@@ -188,10 +200,15 @@ yelp_mallard_document_new (YelpUri *uri)
 {
     YelpMallardDocument *mallard;
     YelpMallardDocumentPrivate *priv;
+    gchar *doc_uri;
 
     g_return_val_if_fail (uri != NULL, NULL);
 
-    mallard = (YelpMallardDocument *) g_object_new (YELP_TYPE_MALLARD_DOCUMENT, NULL);
+    doc_uri = yelp_uri_get_document_uri (uri);
+    mallard = (YelpMallardDocument *) g_object_new (YELP_TYPE_MALLARD_DOCUMENT,
+                                                    "document-uri", doc_uri,
+                                                    NULL);
+    g_free (doc_uri);
     priv = GET_PRIV (mallard);
     priv->uri = g_object_ref (uri);
 
@@ -201,7 +218,6 @@ yelp_mallard_document_new (YelpUri *uri)
     return (YelpDocument *) mallard;
 }
 
-
 static gboolean
 mallard_request_page (YelpDocument         *document,
                       const gchar          *page_id,
@@ -274,7 +290,6 @@ mallard_think (YelpMallardDocument *mallard)
     YelpMallardDocumentPrivate *priv = GET_PRIV (mallard);
     GError *error = NULL;
     YelpDocument *document;
-    gchar **search_path;
     gboolean editor_mode;
 
     gchar **path;
@@ -285,22 +300,20 @@ mallard_think (YelpMallardDocument *mallard)
 
     editor_mode = yelp_settings_get_editor_mode (yelp_settings_get_default ());
 
-    search_path = yelp_uri_get_search_path (priv->uri);
-
-    if (!search_path || search_path[0] == NULL ||
-        !g_file_test (search_path[0], G_FILE_TEST_IS_DIR)) {
+    path = yelp_uri_get_search_path (priv->uri);
+    if (!path || path[0] == NULL ||
+        !g_file_test (path[0], G_FILE_TEST_IS_DIR)) {
         /* This basically only happens when someone passes an actual directory
            manually, which will have a singleton search path.
          */
         error = g_error_new (YELP_ERROR, YELP_ERROR_NOT_FOUND,
                              _("The directory ‘%s’ does not exist."),
-                             search_path[0]);
+                             path[0]);
 	yelp_document_error_pending ((YelpDocument *) document, error);
         g_error_free (error);
 	goto done;
     }
 
-    path = yelp_uri_get_search_path (priv->uri);
     for (path_i = 0; path[path_i] != NULL; path_i++) {
         gfile = g_file_new_for_path (path[path_i]);
         children = g_file_enumerate_children (gfile,
@@ -343,6 +356,14 @@ mallard_think (YelpMallardDocument *mallard)
                 yelp_document_set_page_desc ((YelpDocument *) mallard,
                                              page_data->page_id,
                                              page_data->page_desc);
+                if (page_data->next_page != NULL) {
+                    yelp_document_set_next_id ((YelpDocument *) mallard,
+                                               page_data->page_id,
+                                               page_data->next_page);
+                    yelp_document_set_prev_id ((YelpDocument *) mallard,
+                                               page_data->next_page,
+                                               page_data->page_id);
+                }
                 yelp_document_signal ((YelpDocument *) mallard,
                                       page_data->page_id,
                                       YELP_DOCUMENT_SIGNAL_INFO,
@@ -461,12 +482,15 @@ mallard_page_data_walk (MallardPageData *page_data)
         page_data->xpath = xmlXPathNewContext (page_data->xmldoc);
         mallard_page_data_walk (page_data);
     } else {
+        gboolean ispage;
         xmlNodePtr child, oldcur, oldcache, info;
 
         id = xmlGetProp (page_data->cur, BAD_CAST "id");
         if (id == NULL)
             goto done;
-        if (g_hash_table_lookup (priv->pages_hash, id) != NULL)
+
+        ispage = xml_node_is_ns_name (page_data->cur, MALLARD_NS, BAD_CAST "page");
+        if (ispage && g_hash_table_lookup (priv->pages_hash, id) != NULL)
             goto done;
 
         page_data->cache = xmlNewChild (page_data->cache,
@@ -474,41 +498,15 @@ mallard_page_data_walk (MallardPageData *page_data)
                                         page_data->cur->name,
                                         NULL);
 
-        if (xmlStrEqual (page_data->cur->name, BAD_CAST "page")) {
-            xmlChar *style;
-            gchar **styles;
-            gchar *icon = "help-contents";
+        if (ispage) {
             page_data->page_id = g_strdup ((gchar *) id);
             xmlSetProp (page_data->cache, BAD_CAST "id", id);
             yelp_document_set_page_id ((YelpDocument *) page_data->mallard,
                                        g_strrstr (page_data->filename, G_DIR_SEPARATOR_S),
                                        page_data->page_id);
-            style = xmlGetProp (page_data->cur, BAD_CAST "style");
-            if (style) {
-                gint i;
-                styles = g_strsplit (style, " ", -1);
-                for (i = 0; styles[i] != NULL; i++) {
-                    if (g_str_equal (styles[i], "task")) {
-                        icon = "yelp-page-task";
-                        break;
-                    }
-                    else if (g_str_equal (styles[i], "tip")) {
-                        icon = "yelp-page-tip";
-                        break;
-                    }
-                    else if (g_str_equal (styles[i], "ui")) {
-                        icon = "yelp-page-ui";
-                        break;
-                    }
-                    else if (g_str_equal (styles[i], "video")) {
-                        icon = "yelp-page-video";
-                        break;
-                    }
-                }
-                xmlFree (style);
-            }
             yelp_document_set_page_icon ((YelpDocument *) page_data->mallard,
-                                         page_data->page_id, icon);
+                                         page_data->page_id,
+                                         xml_node_get_icon (page_data->cur));
         } else {
             gchar *newid = g_strdup_printf ("%s#%s", page_data->page_id, id);
             xmlSetProp (page_data->cache, BAD_CAST "id", BAD_CAST newid);
@@ -523,10 +521,10 @@ mallard_page_data_walk (MallardPageData *page_data)
         for (child = page_data->cur->children; child; child = child->next) {
             if (child->type != XML_ELEMENT_NODE)
                 continue;
-            if (xmlStrEqual (child->name, BAD_CAST "info")) {
+            if (xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "info")) {
                 mallard_page_data_info (page_data, child, info);
             }
-            else if (xmlStrEqual (child->name, BAD_CAST "title")) {
+            else if (xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "title")) {
                 xmlNodePtr node;
                 xmlNodePtr title_node = xmlNewChild (page_data->cache,
                                                      priv->cache_ns,
@@ -557,11 +555,11 @@ mallard_page_data_walk (MallardPageData *page_data)
                     xmlXPathObjectPtr obj;
                     page_data->xpath->node = child;
                     obj = xmlXPathCompiledEval (priv->normalize, page_data->xpath);
-                    page_data->page_title = g_strdup (obj->stringval);
+                    page_data->page_title = g_strdup ((const gchar *) obj->stringval);
                     xmlXPathFreeObject (obj);
                 }
             }
-            else if (xmlStrEqual (child->name, BAD_CAST "section")) {
+            else if (xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "section")) {
                 oldcur = page_data->cur;
                 oldcache = page_data->cache;
                 page_data->cur = child;
@@ -588,33 +586,38 @@ mallard_page_data_info (MallardPageData *page_data,
     gboolean editor_mode = yelp_settings_get_editor_mode (yelp_settings_get_default ());
 
     for (child = info_node->children; child; child = child->next) {
-        if (xmlStrEqual (child->name, BAD_CAST "info")) {
+        if (xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "info")) {
             mallard_page_data_info (page_data, child, cache_node);
         }
-        else if (xmlStrEqual (child->name, BAD_CAST "title")) {
+        else if (xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "title")) {
             xmlNodePtr node, title_node;
             xmlChar *type, *role;
             title_node = xmlCopyNode (child, 1);
             xmlAddChild (cache_node, title_node);
 
             type = xmlGetProp (child, BAD_CAST "type");
-            role = xmlGetProp (child, BAD_CAST "role");
 
-            if (xmlStrEqual (type, BAD_CAST "link") && role == NULL)
-                page_data->link_title = TRUE;
-            if (xmlStrEqual (type, BAD_CAST "sort"))
-                page_data->sort_title = TRUE;
-            if (xmlStrEqual (type, BAD_CAST "text")) {
-                YelpMallardDocumentPrivate *priv = GET_PRIV (page_data->mallard);
-                xmlXPathObjectPtr obj;
-                page_data->xpath->node = child;
-                obj = xmlXPathCompiledEval (priv->normalize, page_data->xpath);
-                g_free (page_data->page_title);
-                page_data->page_title = g_strdup (obj->stringval);
-                xmlXPathFreeObject (obj);
+            if (type != NULL) {
+                role = xmlGetProp (child, BAD_CAST "role");
+                if (xmlStrEqual (type, BAD_CAST "link") && role == NULL)
+                    page_data->link_title = TRUE;
+                if (xmlStrEqual (type, BAD_CAST "sort"))
+                    page_data->sort_title = TRUE;
+                if (xmlStrEqual (type, BAD_CAST "text")) {
+                    YelpMallardDocumentPrivate *priv = GET_PRIV (page_data->mallard);
+                    xmlXPathObjectPtr obj;
+                    page_data->xpath->node = child;
+                    obj = xmlXPathCompiledEval (priv->normalize, page_data->xpath);
+                    g_free (page_data->page_title);
+                    page_data->page_title = g_strdup (obj->stringval);
+                    xmlXPathFreeObject (obj);
+                }
+                if (role != NULL)
+                    xmlFree (role);
+                xmlFree (type);
             }
         }
-        else if (xmlStrEqual (child->name, BAD_CAST "desc")) {
+        else if (xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "desc")) {
             YelpMallardDocumentPrivate *priv = GET_PRIV (page_data->mallard);
             xmlXPathObjectPtr obj;
             page_data->xpath->node = child;
@@ -624,10 +627,32 @@ mallard_page_data_info (MallardPageData *page_data,
 
             xmlAddChild (cache_node, xmlCopyNode (child, 1));
         }
-        else if (xmlStrEqual (child->name, BAD_CAST "link")) {
+        else if (xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "link")) {
+            xmlChar *type, *next;
+
+            xmlAddChild (cache_node, xmlCopyNode (child, 1));
+
+            type = xmlGetProp (child, BAD_CAST "type");
+            if (type != NULL) {
+                if (xmlStrEqual (type, "next")) {
+                    next = xmlGetProp (child, BAD_CAST "xref");
+                    if (next != NULL) {
+                        if (page_data->next_page != NULL)
+                            g_free (page_data->next_page);
+                        page_data->next_page = g_strdup (next);
+                        xmlFree (next);
+                    }
+                }
+                xmlFree (type);
+            }
+
+        }
+        else if (xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "revision")) {
             xmlAddChild (cache_node, xmlCopyNode (child, 1));
         }
-        else if (xmlStrEqual (child->name, BAD_CAST "revision")) {
+        else if (xml_node_is_ns_name (child, MALLARD_FACET_NS, BAD_CAST "tag") ||
+                 xml_node_is_ns_name (child, MALLARD_FACET_NS, BAD_CAST "match") ||
+                 xml_node_is_ns_name (child, MALLARD_FACET_NS, BAD_CAST "choice") ) {
             xmlAddChild (cache_node, xmlCopyNode (child, 1));
         }
     }
@@ -685,6 +710,7 @@ mallard_page_data_free (MallardPageData *page_data)
         xmlXPathFreeContext (page_data->xpath);
     g_free (page_data->page_title);
     g_free (page_data->page_desc);
+    g_free (page_data->next_page);
     g_free (page_data);
 }
 
@@ -772,3 +798,268 @@ transform_error (YelpTransform   *transform,
 
     mallard_page_data_cancel (page_data);
 }
+
+static const char *
+xml_node_get_icon (xmlNodePtr node)
+{
+    xmlChar *style;
+    gchar **styles;
+    gchar *icon = "yelp-page";
+    style = xmlGetProp (node, BAD_CAST "style");
+    if (style) {
+        gint i;
+        styles = g_strsplit (style, " ", -1);
+        for (i = 0; styles[i] != NULL; i++) {
+            if (g_str_equal (styles[i], "video")) {
+                icon = "yelp-page-video";
+                break;
+            }
+            else if (g_str_equal (styles[i], "task")) {
+                icon = "yelp-page-task";
+                break;
+            }
+            else if (g_str_equal (styles[i], "tip")) {
+                icon = "yelp-page-tip";
+                break;
+            }
+            else if (g_str_equal (styles[i], "problem")) {
+                icon = "yelp-page-problem";
+                break;
+            }
+            else if (g_str_equal (styles[i], "ui")) {
+                icon = "yelp-page-ui";
+                break;
+            }
+        }
+        g_strfreev (styles);
+        xmlFree (style);
+    }
+    return icon;
+}
+
+static gboolean
+xml_node_is_ns_name (xmlNodePtr      node,
+                     const xmlChar  *ns,
+                     const xmlChar  *name)
+{
+    if (node->ns == NULL)
+        return (ns == NULL);
+    else if (ns != NULL && node->ns->href != NULL)
+        return (xmlStrEqual (ns, node->ns->href) && xmlStrEqual (name, node->name)); 
+    return FALSE;
+}
+
+/******************************************************************************/
+
+typedef struct {
+    YelpMallardDocument *mallard;
+    xmlDocPtr doc;
+    xmlNodePtr cur;
+    GString *str;
+    gboolean is_inline;
+} MallardIndexData;
+
+static void
+mallard_index_node (MallardIndexData *index)
+{
+    xmlNodePtr orig, child;
+    gboolean was_inline;
+
+    orig = index->cur;
+    was_inline = index->is_inline;
+
+    for (child = index->cur->children; child; child = child->next) {
+        if (index->is_inline) {
+            if ((xml_node_is_ns_name (child->parent, MALLARD_NS, BAD_CAST "guiseq") ||
+                 xml_node_is_ns_name (child->parent, MALLARD_NS, BAD_CAST "keyseq")) &&
+                child->prev != NULL) {
+                g_string_append_c (index->str, ' ');
+            }
+            if (child->type == XML_TEXT_NODE) {
+                g_string_append (index->str, child->content);
+                continue;
+            }
+        }
+
+        if (child->type != XML_ELEMENT_NODE ||
+            xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "info") ||
+            xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "comment"))
+            continue;
+
+        if (xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "p") ||
+            xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "code") ||
+            xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "screen") ||
+            xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "title") ||
+            xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "desc") ||
+            xml_node_is_ns_name (child, MALLARD_NS, BAD_CAST "cite")) {
+            index->is_inline = TRUE;
+        }
+
+        index->cur = child;
+        mallard_index_node (index);
+
+        if (index->is_inline && !was_inline) {
+            g_string_append_c (index->str, '\n');
+        }
+
+        index->cur = orig;
+        index->is_inline = was_inline;
+    }
+}
+
+static gboolean
+mallard_index_done (YelpMallardDocument *mallard)
+{
+    g_object_set (mallard, "indexed", TRUE, NULL);
+    return FALSE;
+}
+
+static void
+mallard_index_threaded (YelpMallardDocument *mallard)
+{
+    gchar **path;
+    gint path_i;
+    GHashTable *ids;
+    gchar *doc_uri;
+    YelpMallardDocumentPrivate *priv = GET_PRIV (mallard);
+
+    doc_uri = yelp_uri_get_document_uri (priv->uri);
+    ids = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+    path = yelp_uri_get_search_path (priv->uri);
+    for (path_i = 0; path[path_i] != NULL; path_i++) {
+        GFile *gfile;
+        GFileEnumerator *children;
+        GFileInfo *pageinfo;
+        gfile = g_file_new_for_path (path[path_i]);
+        children = g_file_enumerate_children (gfile,
+                                              G_FILE_ATTRIBUTE_STANDARD_NAME,
+                                              G_FILE_QUERY_INFO_NONE,
+                                              NULL, NULL);
+        while ((pageinfo = g_file_enumerator_next_file (children, NULL, NULL))) {
+            gchar *filename;
+            GFile *pagefile;
+            xmlParserCtxtPtr parserCtxt = NULL;
+            xmlXPathContextPtr xpath = NULL;
+            xmlXPathObjectPtr obj;
+            MallardIndexData *index = NULL;
+            xmlChar *id;
+            YelpUri *uri;
+            gchar *title, *desc, *fulltext, *tmp, *full_uri;
+
+            filename = g_file_info_get_attribute_as_string (pageinfo,
+                                                            G_FILE_ATTRIBUTE_STANDARD_NAME);
+            if (!g_str_has_suffix (filename, ".page")) {
+                g_free (filename);
+                g_object_unref (pageinfo);
+                continue;
+            }
+            pagefile = g_file_resolve_relative_path (gfile, filename);
+            g_free (filename);
+            filename = g_file_get_path (pagefile);
+
+            index = g_new0 (MallardIndexData, 1);
+            index->mallard = mallard;
+
+            parserCtxt = xmlNewParserCtxt ();
+            index->doc = xmlCtxtReadFile (parserCtxt, filename, NULL,
+                                   XML_PARSE_DTDLOAD | XML_PARSE_NOCDATA |
+                                   XML_PARSE_NOENT   | XML_PARSE_NONET   );
+            if (index->doc == NULL)
+                goto done;
+            if (xmlXIncludeProcessFlags (index->doc,
+                                         XML_PARSE_DTDLOAD | XML_PARSE_NOCDATA |
+                                         XML_PARSE_NOENT   | XML_PARSE_NONET   )
+                < 0)
+                goto done;
+
+            index->cur = xmlDocGetRootElement (index->doc);
+
+            id = xmlGetProp (index->cur, BAD_CAST "id");
+            if (id == NULL)
+                goto done;
+            if (g_hash_table_lookup (ids, id))
+                goto done;
+            /* We never use the value. Just need something non-NULL. */
+            g_hash_table_insert (ids, g_strdup (id), id);
+
+            xpath = xmlXPathNewContext (index->doc);
+            xmlXPathRegisterNs (xpath, BAD_CAST "mal", BAD_CAST MALLARD_NS);
+            obj = xmlXPathEvalExpression (BAD_CAST
+                                          "normalize-space((/mal:page/mal:info/mal:title[@type='text'] |"
+                                          "/mal:page/mal:title)[1])",
+                                          xpath);
+            title = g_strdup (obj->stringval);
+            xmlXPathFreeObject (obj);
+            obj = xmlXPathEvalExpression (BAD_CAST
+                                          "normalize-space(/mal:page/mal:info/mal:desc[1])",
+                                          xpath);
+            desc = g_strdup (obj->stringval);
+            xmlXPathFreeObject (obj);
+
+            index->str = g_string_new (desc);
+            g_string_append_c (index->str, ' ');
+            mallard_index_node (index);
+
+            tmp = g_strconcat ("xref:", id, NULL);
+            uri = yelp_uri_new_relative (priv->uri, tmp);
+            yelp_uri_resolve_sync (uri);
+            full_uri = yelp_uri_get_canonical_uri (uri);
+            g_free (tmp);
+            g_object_unref (uri);
+
+            fulltext = g_string_free (index->str, FALSE);
+
+            yelp_storage_update (yelp_storage_get_default (),
+                                 doc_uri, full_uri,
+                                 title, desc,
+                                 xml_node_get_icon (xmlDocGetRootElement (index->doc)),
+                                 fulltext);
+            if (g_str_equal (id, "index"))
+                yelp_storage_set_root_title (yelp_storage_get_default (),
+                                             doc_uri, title);
+            g_free (full_uri);
+            g_free (title);
+            g_free (desc);
+            g_free (fulltext);
+
+        done:
+            if (id)
+                xmlFree (id);
+            if (xpath != NULL)
+                xmlXPathFreeContext (xpath);
+            if (index->doc != NULL)
+                xmlFreeDoc (index->doc);
+            if (index != NULL)
+                g_free (index);
+            if (parserCtxt != NULL)
+                xmlFreeParserCtxt (parserCtxt);
+            g_object_unref (pagefile);
+            g_free (filename);
+            g_object_unref (pageinfo);
+        }
+    }
+    g_strfreev (path);
+    g_hash_table_destroy (ids);
+    priv->index_running = FALSE;
+    g_free (doc_uri);
+    g_object_unref (mallard);
+    g_idle_add ((GSourceFunc) mallard_index_done, mallard);
+}
+
+static void
+mallard_index (YelpDocument *document)
+{
+    YelpMallardDocumentPrivate *priv;
+    gboolean done;
+
+    g_object_get (document, "indexed", &done, NULL);
+    if (done)
+        return;
+
+    priv = GET_PRIV (document);
+    g_object_ref (document);
+    priv->index = g_thread_create ((GThreadFunc) mallard_index_threaded,
+                                   document, FALSE, NULL);
+    priv->index_running = TRUE;
+}
+
